#import "EL2OpenGLView.h"

#import "gfx.h"
#import "TrackRenderer.h"
#import "TrackSynthesizer.h"

#import <fstream>
#import <sstream>
#import <string>
#import <boost/archive/text_iarchive.hpp>
#import <boost/archive/text_oarchive.hpp>
#import <boost/serialization/vector.hpp>


// audio callback: not on the main thread
int audioCallback(void *outputBuffer, void *inputBuffer, unsigned int nFrames, double streamTime, RtAudioStreamStatus status, void *user_data) {
	
	stk::StkFloat *buffy = (stk::StkFloat *)outputBuffer;
	Synthesizer *synthesizer = (Synthesizer *)user_data;
	
	for (unsigned int i=0; i<nFrames; i++) {
		// write the sample
		buffy[i] = synthesizer->tick() * 1.0;
	}
	
	return 0;
}


@implementation EL2OpenGLView

#pragma mark load and save
- (NSString *)projectArchive {
	stringstream s;
	boost::archive::text_oarchive outarchive(s);
	outarchive << m_tracks;
	
	string str = s.str();
	const char *c_str = str.c_str();
	NSString *d = [NSString stringWithCString:c_str encoding:NSUTF8StringEncoding];
	
	return d;
}

- (void)loadProjectArchive:(NSString *)project {
	// load new track
	if (project) {
		string str((const char *)[project UTF8String]);
		
		stringstream s(str);
		boost::archive::text_iarchive inarchive(s);
		
		vector<Track *> *newTracks = NULL;
		inarchive >> newTracks;
		
		// remove all existing tracks, synths, and renderers
		m_synthesizer->removeAllChildren();
		m_renderer->removeAllChildren();
		
		for (unsigned int i=0; i<m_tracks->size(); i++) {
			delete m_tracks->at(i);
			m_tracks->at(i) = NULL;
		}
		m_tracks->clear();
		m_currentTrack = -1;
		
		// add all the new tracks
		for (unsigned int i=0; i<newTracks->size(); i++) {
			[self addTrack:newTracks->at(i)];
		}
	}
}

#pragma mark Input event handlers

-(void)keyDown:(NSEvent *)theEvent {
    NSString *characters = [theEvent characters];
    if ([characters length]) {
        unichar character = [characters characterAtIndex:0];
		int midiVal = -1;
		switch (character) {
			case '+':
			{
				[self addTrack];
				break;
			}
			case NSDeleteFunctionKey:
			case NSDeleteCharFunctionKey:
			{
				// get the track that is to be deleted
				if (m_currentTrack < m_tracks->size() && m_currentTrack >= 0) {
					// remove track renderer for old track
					for (unsigned int i=0; i<m_renderer->numChildren(); i++) {
						if(TrackRenderer *tr = dynamic_cast<TrackRenderer *>(m_renderer->childAt(i))) {
							if (tr->track() == m_tracks->at(m_currentTrack)) {
								
								m_renderer->removeChild(tr);
								
								// adjust position of remaining tracks
								for (int r=i-1; r>=0; r--) {
									Renderer *c = m_renderer->childAt(r);
									c->setPosition(c->xPosition(), c->yPosition() - 100.0f, c->zPosition());
								}
								
								break;
							}
						}
					}
					
					// remove track synth for old track
					for (unsigned int i=0; i<m_synthesizer->numChildren(); i++) {
						if(TrackSynthesizer *ts = dynamic_cast<TrackSynthesizer *>(m_synthesizer->childAt(i))) {
							if (ts->track() == m_tracks->at(m_currentTrack)) {
								m_synthesizer->removeChild(ts);
								break;
							}
						}
					}
					
					// delete old track
					delete m_tracks->at(m_currentTrack);
					m_tracks->at(m_currentTrack) = NULL;
					m_tracks->erase(m_tracks->begin()+m_currentTrack);
					
					// update m_currentTrack
					if (m_currentTrack == 0 && m_tracks->size() > 0) {
						[self setCurrentTrack:0];
					} else {
						[self setCurrentTrack:m_currentTrack-1];
					}
				}
				
				break;
			}
			case NSUpArrowFunctionKey:
			{
				if (m_currentTrack < m_tracks->size() - 1) {
					[self setCurrentTrack:m_currentTrack + 1];
				}
				break;
			}
			case NSDownArrowFunctionKey:
			{
				if (m_currentTrack > 0) {
					[self setCurrentTrack:m_currentTrack - 1];
				}
				break;
			}
			case NSRightArrowFunctionKey:
			{
				if (m_tracks->size() > 0) {
					m_tracks->at(m_currentTrack)->setInstrumentID((m_tracks->at(m_currentTrack)->instrumentID() + 1) % 2);
				}
				break;
			}
			case '`':
				midiVal = 39;
				break;
			case 'q':
				midiVal = 41;
				break;
			case '2':
				midiVal = 42;
				break;
			case 'w':
				midiVal = 43;
				break;
			case '3':
				midiVal = 44;
				break;
			case 'e':
				midiVal = 45;
				break;
			case '4':
				midiVal = 46;
				break;
			case 'r':
				midiVal = 47;
				break;
			case 't':
				midiVal = 48;
				break;
			case '6':
				midiVal = 49;
				break;
			case 'y':
				midiVal = 50;
				break;
			case '7':
				midiVal = 51;
				break;
			case 'u':
				midiVal = 52;
				break;
			case 'i':
				midiVal = 53;
				break;
			case '9':
				midiVal = 54;
				break;
			case 'o':
				midiVal = 55;
				break;
			case '0':
				midiVal = 56;
				break;
			case 'p':
				midiVal = 57;
				break;
			case '-':
				midiVal = 58;
				break;
			case '[':
				midiVal = 59;
				break;
			case ']':
				midiVal = 60;
				break;
			case 'z':
				midiVal = 60;
				break;
			case 's':
				midiVal = 61;
				break;
			case 'x':
				midiVal = 62;
				break;
			case 'd':
				midiVal = 63;
				break;
			case 'c':
				midiVal = 64;
				break;
			case 'v':
				midiVal = 65;
				break;
			case 'g':
				midiVal = 66;
				break;
			case 'b':
				midiVal = 67;
				break;
			case 'h':
				midiVal = 68;
				break;
			case 'n':
				midiVal = 69;
				break;
			case 'j':
				midiVal = 70;
				break;
			case 'm':
				midiVal = 71;
				break;
			case ',':
				midiVal = 72;
				break;
			case 'l':
				midiVal = 73;
				break;
			case '.':
				midiVal = 74;
				break;
			case ';':
				midiVal = 75;
				break;
			case '/':
				midiVal = 76;
				break;
		}
		
		if (midiVal != -1 && m_tracks->size() > 0) {
			m_tracks->at(m_currentTrack)->addNote(m_tracks->at(m_currentTrack)->nextSample() + stk::RT_BUFFER_SIZE,
												  m_tracks->at(m_currentTrack)->nextSample() + stk::RT_BUFFER_SIZE + 1000,
												  8.1757989156 * pow(2.0,((float)midiVal)/12.0),
												  1.0);
		}
		
		[self setNeedsDisplay:YES];
	}
}

#pragma mark drawing event handlers

- (void)prepareOpenGL {
//    GLint swapInt = 1;
//    [[self openGLContext] setValues:&swapInt forParameter:NSOpenGLCPSwapInterval]; // set to vbl sync
	
	// init GL stuff here
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	glShadeModel(GL_SMOOTH);
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glBlendEquation(GL_ADD);
	
	glClearColor(0.95f, 0.95f, 0.95f, 1.0f);
	
	[self updateMatrices];
	[self addTrack];
}

- (void)updateMatrices {
	// setup viewport and prespective if necessary
	if (m_size.height != self.bounds.size.height || m_size.width != self.bounds.size.width) {
		
		// update m_size
		m_size.height = self.bounds.size.height;
		m_size.width = self.bounds.size.width;
		
		[[self openGLContext] makeCurrentContext];
		
		// map the view port to the client area
		glViewport(0, 0, m_size.width, m_size.height);
		
		// set up an orthographic projection
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0.0f, m_size.width, m_size.height, 0.0f);
		
		// set an identity modelview matrix
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	}
}

- (void)animationTimer:(NSTimer *)theTimer {
	[self setNeedsDisplay:YES];
}

- (void)drawRect:(NSRect)rect {
	[super drawRect:rect];
	
	[self updateMatrices];
	
    [[self openGLContext] makeCurrentContext];
	
	// clear our drawable
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// draw
	XGfx::getCurrentTime(true);
	
	// draw the scene
	m_renderer->interpolate(XGfx::delta());
	m_renderer->render();
	
	[[self openGLContext] flushBuffer];
}

#pragma mark helpers

- (void)addTrack {
	[self addTrack:NULL];
}

- (void)addTrack:(Track *)t {
	if (t) {
		m_tracks->push_back(t);
	} else {
		m_tracks->push_back( new Track(stk::Stk::sampleRate() * 4, 0) );
	}
	
	// sync next sample
	m_tracks->back()->setNextSample(m_tracks->front()->nextSample());
	
	// update m_currentTrack
	[self setCurrentTrack:(m_tracks->size() - 1)];
	
	TrackRenderer *r = new TrackRenderer(m_tracks->back());
	r->setPosition(0.0f, -100.0f, 0.0f);
	r->setWidth(m_size.width);
	r->setHeight(100.0f);
	r->setColor(0.5f, 0.5f, 0.5f);
	r->setAlpha(0.4f);
	r->setSelectedColor(0.2f, 0.4f, 0.6f);
	r->setSelectedAlpha(0.9);
	r->interpolate(1.0);
	
	r->setPosition(0.0f, 0.0f, 0.0f);
	
	// update positions of existing tracks
	for (unsigned int i=0; i<m_renderer->numChildren(); i++) {
		Renderer *c = m_renderer->childAt(i);
		c->setPosition(0.0f, c->yPosition() + 100.0f, 0.0f);
	}
	
	m_renderer->addChild(r);
	
	TrackSynthesizer *s = new TrackSynthesizer(m_tracks->back());
	m_synthesizer->addChild(s);
}

- (void)setCurrentTrack:(int)i {
	if (i < m_tracks->size()) {
		
		// deselect previous current track
		if (m_currentTrack >= 0 && m_currentTrack < m_tracks->size()) {
			m_tracks->at(m_currentTrack)->deselect();
		}
		
		// change m_currentTrack
		m_currentTrack = i;
		
		// select new current track
		if (m_currentTrack >= 0) {
			m_tracks->at(m_currentTrack)->select();
		}
	}
}

#pragma mark Initialization, etc.

-(void)awakeFromNib {
	// init m_size
	m_size = CGSizeZero;
	
	// start animation timer
	timer = [NSTimer timerWithTimeInterval:(1.0f/60.0f) target:self selector:@selector(animationTimer:) userInfo:nil repeats:YES];
	[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSDefaultRunLoopMode];
	[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSEventTrackingRunLoopMode]; // ensure timer fires during resize
	[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSModalPanelRunLoopMode];
	
	// the global track
	m_currentTrack = -1;

	// STK setup
	NSString *resourcesPath = [[NSBundle mainBundle] resourcePath];
	stk::Stk::setRawwavePath( [resourcesPath UTF8String] );
	
	// initialize the root renderer and synthesizer
	m_renderer = new Renderer();
	m_synthesizer = new Synthesizer();
	
	// initialize the track and a render and synthesizer for it
	m_tracks = new vector<Track *>;
	
	// RtAudio setup
	m_audio = NULL;	// RtAudio pointer
	
	// create the RtAudio and start the audio stream
	try {
		m_audio = new RtAudio::RtAudio();
		
		// output params
		RtAudio::StreamParameters outputParams;
		outputParams.deviceId = m_audio->getDefaultOutputDevice();;
		outputParams.nChannels = 1;
		
		// format
		RtAudioFormat format = (sizeof(stk::StkFloat) == 8) ? RTAUDIO_FLOAT64 : RTAUDIO_FLOAT32;
		
		// nFrames
		unsigned int bufferFrames = stk::RT_BUFFER_SIZE;
		
		// options
		RtAudio::StreamOptions options;
		options.numberOfBuffers = 8;
		options.streamName = ::string("everylayer");
		
		m_audio->openStream(&outputParams,						// output parameters
						  NULL,									// input parameters
						  format,								// format
						  (unsigned int)stk::Stk::sampleRate(), // sampleRate
						  &bufferFrames,						// nFrames
						  &audioCallback,						// callback
						  m_synthesizer,						// userdata
						  &options);							// options
		m_audio->startStream();
		
	} catch (RtError &error) {
		error.printMessage();
		[[NSApplication sharedApplication] performSelector:@selector(terminate:) withObject:nil afterDelay:0.0];
	}
}

- (BOOL)acceptsFirstResponder {
  return YES;
}

- (BOOL)becomeFirstResponder {
  return  YES;
}

- (BOOL)resignFirstResponder {
  return YES;
}

- (void)dealloc {
	
	[timer invalidate];
	[timer release];
	timer = nil;
	
	if (m_synthesizer)	{ delete m_synthesizer; m_synthesizer	= NULL; }
	if (m_renderer)		{ delete m_renderer;	m_renderer		= NULL; }
	if (m_tracks) {
		for (unsigned int i=0; i<m_tracks->size(); i++) {
			delete m_tracks->at(i);
			m_tracks->at(i) = NULL;
		}
		
		delete m_tracks;
		m_tracks = NULL;
	}
	if (m_audio) {
		m_audio->closeStream();
		delete m_audio;
		m_audio = NULL;
	}
	
	[super dealloc];
}

@end
